#########################################################################################
## ACKNOWLEDGEMENTS
## the Wholesale Customers dataset that is used in this example is available from 
## the UCI ML Repository: https://archive.ics.uci.edu/ml/datasets/Wholesale+customers
#########################################################################################

customers.data <- read.csv(file = "data/wholesale_customers_data.csv")
str(customers.data)

## Channel and Region are nominal (factor) variables
## with the following set of possible values:
## Region
## 1 - Lisbon
## 2 - Oporto
## 3 - Other_region
customers.data$Region <- factor(customers.data$Region, levels = c(1,2,3),
                                labels = c('Lisbon', 'Oporto', 'Other_region'))
## Channel:
## 1 - Horeca (Hotel/Restaurant/Cafee)
## 2 - Retail
customers.data$Channel <- factor(customers.data$Channel, levels = c(1,2),
                                 labels = c('Horeca', 'Retail'))

## the rest of the attributes represent customer's annual spending on different 
## kinds of products (fresh products, dairy products, ...)

##################################
# EXAMINING AND PREPARING THE DATA
##################################

## check if there are observations with missing values
which(complete.cases(customers.data)==F)
## no missing values

## k-means tends to perform poorly when outliers are present;  
## so, we should check if outliers are present
## we're check only numerical variables that will be used for clustering
summary(customers.data[,c(3:8)])
## it seems that all 6 variables have outliers 
## box-plots are useful for detection of outliers
## check: https://flowingdata.com/2008/02/15/how-to-read-and-use-a-box-and-whisker-plot/
library(ggplot2)
ggplot(customers.data, aes(x=Channel, y=Grocery, fill=Channel)) + geom_boxplot()
ggplot(customers.data, aes(x=Channel, y=Milk, fill=Channel)) + geom_boxplot()
ggplot(customers.data, aes(x=Channel, y=Delicassen, fill=Channel)) + geom_boxplot()
ggplot(customers.data, aes(x=Channel, y=Frozen, fill=Channel)) + geom_boxplot()
ggplot(customers.data, aes(x=Channel, y=Fresh, fill=Channel)) + geom_boxplot()
ggplot(customers.data, aes(x=Channel, y=Detergents_Paper, fill=Channel)) + geom_boxplot()

## obviously outliers are present, but the plots also suggest that that there is a 
## considerable difference between the two distribution channels, 
## so it would be better to examine and cluster each of them separately 
retail.data <- subset(customers.data, Channel == 'Retail')
horeca.data <- subset(customers.data, Channel == 'Horeca')

## let's start by focusing on the retail.data
## for homework: do the same with the horeca.data

summary(retail.data)
## remove the Channel variable as we now have just one channel
retail.data <- retail.data[,-1]

## as a way of dealing with outliers, we'll use the Winsorizing technique
## https://en.wikipedia.org/wiki/Winsorizing
## Practically, it consists of replacing extreme values with a specific 
## percentile of the data, typically 90th or 95th 

## let's start with the Grocery variable
boxplot(retail.data$Grocery, xlab='Grocery')
## look at the outliers more closely
sort(boxplot.stats(retail.data$Grocery)$out)
## examining the 90th, 95th, ... percentile
quantile(retail.data$Grocery, probs = seq(from = 0.9, to = 1, by = 0.025))
## the 95th percentile seems to be a good cutting point
new.max <- as.numeric(quantile(retail.data$Grocery, probs = 0.95))
retail.data$Grocery[retail.data$Grocery > new.max] <- new.max
## check if outliers are still present
boxplot(retail.data$Grocery, xlab='Grocery')
## it's fine now

## the Fresh variable
boxplot(retail.data$Fresh, xlab='Fresh')
sort(boxplot.stats(retail.data$Fresh)$out)
quantile(retail.data$Fresh, probs = seq(0.9, 1, 0.025))
new.max <- as.numeric(quantile(retail.data$Fresh, probs = 0.95))
retail.data$Fresh[retail.data$Fresh > new.max] <- new.max
boxplot(retail.data$Fresh, xlab='Fresh')

## do the same for the Delicassen variable
boxplot(retail.data$Delicassen, xlab='Delicassen')
sort(boxplot.stats(retail.data$Delicassen)$out)
quantile(retail.data$Delicassen, probs = seq(0.9, 1, 0.025))
## setting values to the 90th percentile seems to be a good approach
new.max <- as.numeric(quantile(retail.data$Delicassen, probs = 0.925))
retail.data$Delicassen[retail.data$Delicassen > new.max] <- new.max
## check if outliers are still present
boxplot(retail.data$Delicassen, xlab='Delicassen')

## the Frozen variable
boxplot(retail.data$Frozen, xlab='Frozen')
sort(boxplot.stats(retail.data$Frozen)$out)
quantile(retail.data$Frozen, probs = c(seq(0.9, 1, 0.025)))
## setting values to the 93rd percentile seems to be a good approach
new.max <- as.numeric(quantile(retail.data$Frozen, probs = 0.925))
retail.data$Frozen[retail.data$Frozen > new.max] <- new.max
## check if outliers are still present
boxplot(retail.data$Frozen, xlab='Frozen')

## the Milk variable
boxplot(retail.data$Milk, xlab='Milk')
sort(boxplot.stats(retail.data$Milk)$out)
quantile(retail.data$Milk, probs = seq(0.9, 1, 0.025))
new.max <- as.numeric(quantile(retail.data$Milk, probs = 0.90))
retail.data$Milk[retail.data$Milk > new.max] <- new.max
boxplot(retail.data$Milk, xlab='Milk')

## the Detergents_Paper variable
boxplot(retail.data$Detergents_Paper, xlab='Detergents and Paper')
sort(boxplot.stats(retail.data$Detergents_Paper)$out)
quantile(retail.data$Detergents_Paper, probs = seq(0.9, 1, 0.025))
new.max <- as.numeric(quantile(retail.data$Detergents_Paper, probs = 0.925))
retail.data$Detergents_Paper[retail.data$Detergents_Paper > new.max] <- new.max
boxplot(retail.data$Detergents_Paper, xlab='Detergents and Paper')

summary(retail.data)

#############################
# CLUSTERING WITH 2 FEATURES
#############################

## choose features for this inital, simple clustering
pairs(~Fresh+Frozen+Grocery+Milk+Delicassen+Detergents_Paper,
      data = retail.data)
## no particular pattern can be observed in any of the plots,
## but for the sake of an example, let's pick Frozen and Milk variables

## Get an closer look at selected variables
gg1 <- ggplot(data=retail.data, aes(x=Frozen, y=Milk))
gg1 <- gg1 + geom_point(shape=1)
gg1 <- gg1 + xlab("Annual spending on frozen products")
gg1 <- gg1 + ylab("Annual spending on dairy products")
gg1 <- gg1 + ggtitle("Retail customers annual spending")
gg1
## No clear pattern in the data, but let's run k-means and see if some clusters will emerge

## create a subset of the original data containing the attributes to be used in the K-Means
retail.2f.data <- retail.data[, c("Frozen", "Milk")]
summary(retail.2f.data)
## the two variables have different value ranges, so, we need to normalize the data
## normalization is done using the formula: (x-min(x))/(max(x)-min(x))
retail.2f.norm <- as.data.frame(apply(retail.2f.data, 2, normalize.feature))
summary(retail.2f.norm)

## run the K Means algorithm, specifying, for example, 4 centers
set.seed(3108)
simple.4k <- kmeans(x = retail.2f.norm, centers=4, iter.max=20, nstart=1000)
## <! explain the meaning of iter.max (a 'safenet' in case of slow-to-converge cases) 
## and nstart (multiple instatiations) !> 

## let's inspect the results
simple.4k

## <! explain clustering evaluation metrics:
# - within_SS - within cluster sum of squares, i.e., sum of squared differences between individual
#   data points in a cluster and the cluster center; it is computed for each cluster
# - total_SS: the sum of squared differences of each data point to the global sample mean
# - between_SS: the sum of squared differences of each cluster center to the global sample mean
#   (when computing this value, the squared difference of each cluster center to the global 
#   sample mean is multiplied by the number of data points in that cluster)
# - between_SS / totat_SS: this ratio indicates how 'well' the sample splits into clusters;
#   the higher the ratio, the better clustering
## !>

## add the vector of clusters to the data frame
retail.2f.norm$cluster <- factor(simple.4k$cluster)
head(retail.2f.norm)

## color the points by their respective clusters
gg2 <- ggplot(data=retail.2f.norm, aes(x=Frozen, y=Milk, colour=cluster))
gg2 <- gg2 + geom_point()
gg2 <- gg2 + xlab("Annual spending on frozen products")
gg2 <- gg2 + ylab("Annual spending on dairy products")
gg2 <- gg2 + ggtitle("Retail customers annual spending")
gg2 <- gg2 + scale_fill_brewer(palette="Set1", name="Cluster")
gg2 <- gg2 + theme_bw()
gg2
## add cluster centers
gg2 + geom_point(data=as.data.frame(simple.4k$centers),
                     colour="black",size=4, shape=17)

## the plots allow us to visualy inspect the clusters, but it is better to have a more 
## systematic approach to judging the quality of clustering and selecting the best value for K

#################################
# SELECTING THE BEST VALUE FOR K
#################################

## SELECTING K USING THE ELBOW METHOD
# instead of guessing the correct value for K, we can take a more 
# systematic approach to choosing the 'right' K; it is called the Elbow method, and
# is based on the sum of squared differences between data points and cluster centres,
# that is, the sum of within_SS for all the clusters (tot.withinss) 

## let's compute the metric required for the Elbow method (tot.withinss);
## along the way, we'll also compute the other metric: ratio of between_SS and totat_SS
eval.metrics.2var <- data.frame()
retail.2f.norm <- retail.2f.norm[, c(1,2)] # remove the column with clusters

## run kmeans for all K values in the range 2:8
for(k in 2:8){
  set.seed(3108)
  km.res <- kmeans(x=retail.2f.norm, centers=k, iter.max=20, nstart = 1000)
  # combine cluster number and the error measure, write to df
  eval.metrics.2var <- rbind(eval.metrics.2var, 
                             c(k, km.res$tot.withinss, km.res$betweenss/km.res$totss)) 
}
names(eval.metrics.2var) <- c("cluster", "tot.within.ss", "ratio")
eval.metrics.2var

## draw the Elbow plot
ggplot(data=eval.metrics.2var, aes(x=cluster, y=tot.within.ss, group=1)) + 
  theme_bw() + 
  geom_line(colour = "darkgreen") +
  theme(text = element_text(size=15)) +
  ggtitle("Reduction in error for different values of K\n") +
  xlab("\nClusters") + 
  ylab("Total Within Cluster Sum of Squares\n") +
  scale_x_continuous(breaks=seq(from=0, to=8, by=1))
## it seams that K=3 or K=4 would be the best options for the number of clusters

## if it is not fully clear from the plot where we have significant decrease in the 
## tot.within.ss value, we can compute the difference between each two subsequent values
## the compute.difference is a utility function (given at the end of the script)
data.frame(K=2:8, 
           tot.within.ss.delta=compute.difference(eval.metrics.2var$tot.within.ss),
           ratio.delta=compute.difference(eval.metrics.2var$ratio))


## as we've already examined solution with 4 clusters, let's also examine the solution with K=3
set.seed(3108)
simple.3k <- kmeans(x = retail.2f.norm, centers=3, iter.max=20, nstart=1000)
retail.2f.norm$cluster <- factor(simple.3k$cluster)
## first, plot it
ggplot(data=retail.2f.norm, aes(x=Frozen, y=Milk, colour=cluster)) + 
    geom_point() +
    xlab("Annual spending on frozen products") +
    ylab("Annual spending on dairy products") +
    ggtitle("Retail customers annual spending") +
    scale_fill_brewer(palette="Set1", name="Cluster") +
    theme_bw() + 
    geom_point(data=as.data.frame(simple.3k$centers),
                     colour="black",size=4, shape=17)

## next, we examine clusters closer by looking into the cluster centers (mean) 
## and standard deviation from the centers 
## when examining clusters - in order to interpret them - we use 'regular' (not normalized) features
sum.stats <- summary.stats(feature.set = retail.2f.data, 
                           clusters = simple.3k$cluster, cl.num = 3)
# this is for generating nice table layout
# install.packages("knitr")
library(knitr)
kable(x = sum.stats, format = "rst")
## the solution with K=4 seems to be better as features are less disperesed than here
## (e.g. the Milk variable in cluster 3, and Frozen variable in cluster 2 have very 
## high dispersion); solution with K=4 resolves this (check the plot, and examine cluster centers)

#########################################
# CLUSTERING WITH ALL (NUMERIC) FEATURES
#########################################

## since the 6 numeric attributes differ in their value ranges,
## we need to normalize them, that is, transform them to the value range [0,1]
retail.norm <- as.data.frame(apply(retail.data[,c(2:7)], 2, normalize.feature)) 
summary(retail.norm)

## instead of guessing K, we'll right away use the Elbow method 
## to find the optimal value for K
eval.metrics.6var <- data.frame()
for(k in 2:8){
  set.seed(3108)
  km.res <- kmeans(x=retail.norm, centers=k, iter.max=20, nstart = 1000)
  # combine cluster number and the error measure, write to df
  eval.metrics.6var <- rbind(eval.metrics.6var, 
                             c(k, km.res$tot.withinss, km.res$betweenss/km.res$totss)) 
}
names(eval.metrics.6var) <- c("cluster", "tot.within.ss", "ratio")
eval.metrics.6var

## draw the Elbow plot
ggplot(data=eval.metrics.6var, aes(x=cluster, y=tot.within.ss, group=1)) + 
  theme_bw() + 
  geom_line(colour = "darkgreen") +
  theme(text = element_text(size=15)) +
  ggtitle("Reduction in error for different values of K\n") +
  xlab("\nClusters") + 
  ylab("Total Within Cluster Sum of Squares\n") +
  scale_x_continuous(breaks=seq(from=0, to=8, by=1))
## this time it seems that solution with 3 clusters might be the best; alternatively, 
## the one with 5 clusters

## as previously done, we'll also look for the best value for K by  
## examining the difference between each two subsequent values of
## tot.within.ss and ratio (of between.SS and total.SS):
data.frame(k=c(2:8),
           tot.within.ss.delta=compute.difference(eval.metrics.6var$tot.within.ss),
           ration.delta=compute.difference(eval.metrics.6var$ratio))
## this again suggest 3 clusters as the best option; though it might be worth checking 
## K=5, as well

## USE R's SUPPORT FOR SELECTING THE BEST VALUE FOR K
## there is also an R function that checks numerous of criteria (up to 30) to 
## determine the best number of clusters
# install.packages("NbClust")
require(NbClust)
set.seed(3108)
nc <- NbClust(data = retail.norm, min.nc = 3, max.nc = 6,  method = "kmeans")
## number of criteria that 'voted' for certain number of clusters
table(nc$Best.n[1,])
## this method suggest K=3 as the best option, though it is closely followed by other K values

## examine in detail clustering with K=3
set.seed(3108)
retail.3k.6var <- kmeans(x=retail.norm, centers=3, iter.max=15, nstart = 1000)
## examine cluster centers
sum.stats <- summary.stats(feature.set = retail.norm, #retail.data[,c(2:7)], 
                           clusters = retail.3k.6var$cluster, cl.num = 3)
kable(x = sum.stats, format = "rst")

## check also the solution with K=5 (among the best runner-ups)
set.seed(3108)
retail.5k.6var <- kmeans(x=retail.norm, centers=5, iter.max=15, nstart = 1000)
## examine cluster centers
sum.stats <- summary.stats(feature.set = retail.norm, 
                           clusters = retail.5k.6var$cluster, cl.num = 5)
kable(x = sum.stats, format = "rst")

## examine the distribution of clusters across the regions
retail.data$cl5 <- retail.5k.6var$cluster
region.vs.clusters <- xtabs(~Region+cl5, data = retail.data)
round(region.vs.clusters / rowSums(region.vs.clusters), digits = 2)
## noting particularly interesting here

##############################################
# COMPARISON OF DIFFERENT CLUSTERING SOLUTIONS
##############################################

## now we have a couple of possible clustring solutions, but the question is which 
## is the best one: the clustering model built with 2 or 6 variables, 3 or 5 clusters? 
## let's compare the clustering results produced by these different models
## to that end, we'll use the cluster.stats() f. (from fpc package) - it allows for
## comparing two or more clustering solutions, regardless of the clustering method
## applied to obtain those clusters
install.packages("fpc")
library(fpc)

## let's define the criteria to be used for comparison
?cluster.stats
comparison.criteria <- c("max.diameter", "min.separation", "average.between", 
                         "average.within", "within.cluster.ss")  
d <- dist(x = retail.norm)
comparison <- sapply(list(km.3k.2var=simple.3k$cluster, # model with 2 variables and 3 clusters
                          km.4k.2var=simple.4k$cluster, # model with 2 variables and 4 clusters
                          km.3k.6var=retail.3k.6var$cluster, # model with 6 variables and 3 clusters
                          km.5k.6var=retail.5k.6var$cluster), # model with 6 variables and 5 clusters
                    FUN=function(x)
                          cluster.stats(d, x))[comparison.criteria,]
kable(x = comparison, format = "rst")
## the metrics indicate that the solution with 6 variables and 5 clusters (km.5k.6var) is 
## the best among the four examined models, for the following reasons
## - max.diameter (lower values better): km.5k.6var has the lowest value
## - min.separation (higher values better): km.5k.6var is lagging behind the first 2 models
## - average.between (higher values better): km.5k.6var is underperforming here, 
##                                            but it doesn't differ much from the other models
## - average.within (lower values better): km.5k.6var has the lowest value
## - within.cluster.ss (lower values better): km.5k.6var has the lowest value

######################
## UTILITY FUNCTIONS #
######################

## function that computes the difference between two subsequent values
compute.difference <- function(values) {
  dif <- vector(mode = "numeric", length = length(values))
  dif[1] <- NA
  for(i in 1:(length(values)-1)) {
    dif[i+1] <- abs(values[i+1] - values[i])
  }
  dif
}

## function that provides summary statistics about clusters
summary.stats <- function(feature.set, clusters, cl.num) {
  sum.stats <- aggregate(x = feature.set, 
                         by = list(clusters), 
                         FUN = function(x) { 
                           m <- mean(x, na.rm = T)
                           sd <- sqrt(var(x, na.rm = T))
                           paste(round(m, digits = 2), " (", 
                                 round(sd, digits = 2), ")", sep = "")
                         })
  sum.stat.df <- data.frame(cluster = sum.stats[,1], 
                            freq = as.vector(table(clusters)),
                            sum.stats[,-1])
  
  sum.stats.transpose <- t( as.matrix(sum.stat.df) )
  sum.stats.transpose <- as.data.frame(sum.stats.transpose)
  attributes <- rownames(sum.stats.transpose)
  sum.stats.transpose <- as.data.frame( cbind(attributes, sum.stats.transpose) )
  colnames(sum.stats.transpose) <- c( "attributes", rep("Mean (SD)", cl.num) )
  rownames(sum.stats.transpose) <- NULL
  sum.stats.transpose
}

normalize.feature <- function( feature ) {
  if ( sum(feature, na.rm = T) == 0 ) feature
  else ((feature - min(feature, na.rm = T))/(max(feature, na.rm = T) - min(feature, na.rm = T)))
}

